# TODO this should not be needed in the future
name_to_id_map = {
    'NOWPreprocessor': '2hgojz3z',
    'NOWElasticIndexer': 'mdtuyaaj',
    'NOWAutoCompleteExecutor2': 'w5w084h7',
}

id_to_name_map = {val: key for key, val in name_to_id_map.items()}
